#include "GaussianBlur.h"
#include <math.h>

const float PI=3.14159265358979323846;

using namespace std;

GaussianBlur::GaussianBlur(PClip clip,float sdy,float sdc,int border,bool integrate,float unsharpstrength,int y,int u,int v,IScriptEnvironment* env):
GenericVideoFilter(clip),Y(y),U(u),V(v),gbY(0),gbC(0)
{
	
if(Y==3)
	gbY=new GausBlur(vi.width,vi.height,sdy,border,integrate,unsharpstrength,env);
if(U==3||V==3)
	gbC=new GausBlur(vi.width/2,vi.height/2,sdc,border,integrate,unsharpstrength,env);
}

GaussianBlur::~GaussianBlur()
{
delete gbY;
delete gbC;
}

PVideoFrame GaussianBlur::GetFrame(int n, IScriptEnvironment* env){

	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);
		
	if(Y==3)
		gbY->Blur(PLANAR_Y,src,dst,env);
	else if(Y==2)
		env->BitBlt(dst->GetWritePtr(PLANAR_Y), dst->GetPitch(PLANAR_Y), src->GetReadPtr(PLANAR_Y), src->GetPitch(PLANAR_Y),src->GetRowSize(PLANAR_Y), src->GetHeight(PLANAR_Y));
	else if(Y<1&&Y>-256)
		memset (dst->GetWritePtr(PLANAR_Y), -Y, dst->GetPitch(PLANAR_Y)*dst->GetHeight(PLANAR_Y));

	if(U==3)
		gbC->Blur(PLANAR_U,src,dst,env);
	else if(U==2)
		env->BitBlt(dst->GetWritePtr(PLANAR_U), dst->GetPitch(PLANAR_U), src->GetReadPtr(PLANAR_U), src->GetPitch(PLANAR_U),src->GetRowSize(PLANAR_U), src->GetHeight(PLANAR_U));
	else if(U<1&&U>-256)
		memset (dst->GetWritePtr(PLANAR_U), -U, dst->GetPitch(PLANAR_U)*dst->GetHeight(PLANAR_U));

	if(V==3)
		gbC->Blur(PLANAR_V,src,dst,env);
	else if(V==2)
		env->BitBlt(dst->GetWritePtr(PLANAR_V), dst->GetPitch(PLANAR_V), src->GetReadPtr(PLANAR_V), src->GetPitch(PLANAR_V),src->GetRowSize(PLANAR_V), src->GetHeight(PLANAR_V));
	else if(V<1&&V>-256)
		memset (dst->GetWritePtr(PLANAR_V), -V, dst->GetPitch(PLANAR_V)*dst->GetHeight(PLANAR_V));
	return dst;
}


GausBlur::GausBlur(unsigned int _width,unsigned int _height,float sd,int _border,bool integrate,float unsharpstrength, IScriptEnvironment* env):
sd(sd),border(_border),srcheight(height),srcwidth(width),outwidth(width/2+1),in(0),out(0),realout(0),gaus(0),norm(1/(width*height)),radius(sd*3.0+0.5),width(_width),height(_height)
{
		if(border<-255)
			border=-255;
		if(border>3)
			border=3;

		if(radius>=height)
			radius=height-1;
		if(radius>=width)
			radius=width-1;

		if(border>2)
		{
			width+=radius*2;
			width=(width+31)/32*32;
			height+=radius*2;
			height=(height+31)/32*32;
			outwidth=width/2+1;
		}
		else if(border<1)
		{
			width+=radius;
			width=(width+31)/32*32;
			height+=radius;
			height=(height+31)/32*32;
			outwidth=width/2+1;
		}
		else if(border==1)
			radius=min(width,height)/2;

		in=new float[height*width];
		realout=new float[height*width];
		out= new complex<float>[height*outwidth];
		gaus=new complex<float>[height*outwidth];
		//fftwf_complex* o1=( fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex) * height*outwidth);
		//float *i1 = (float *)fftwf_malloc(sizeof(float) * width*height);
		plan = fftwf_plan_dft_r2c_2d(height, width, in, reinterpret_cast<fftwf_complex *>(out), FFTW_MEASURE); // direct fft 
		//plan = fftwf_plan_dft_r2c_2d(1024, 512, i1,o1, FFTW_MEASURE); // direct fft 
		plani = fftwf_plan_dft_c2r_2d(height, width, reinterpret_cast<fftwf_complex *>(out), realout, FFTW_MEASURE|FFTW_DESTROY_INPUT); // inverse fft
		norm=1.0/(width* height);

		CalcGaus(integrate,unsharpstrength!=0.0,unsharpstrength);
		if(border<1)
			for(unsigned int i=0;i<width*height;i++)
				in[i]=abs(border);
		
		env->BitBlt((BYTE*)gaus,outwidth*sizeof(complex<float>),(BYTE*)out,outwidth*sizeof(complex<float>),outwidth*sizeof(complex<float>),height);
}

GausBlur::~GausBlur()
{
	delete[] in;
	delete[] out;
	delete[] realout;
	delete[] gaus;
	fftwf_destroy_plan(plan);
	fftwf_destroy_plan(plani);
}

void GausBlur::CalcGaus(bool integrate,bool unsharp,float strength)
{
	SecureZeroMemory(in,height*width*sizeof(float));
	float totalsum=0;
	

	int cellcount=(2*radius+1)*(2*radius+1);
	for(unsigned int y=0,offset=0;y<radius;y++,offset+=width)
	{
		if(integrate)
		{
			float yint=RombergIntegral(y-0.5,y+0.5,sd);
			for(unsigned int x=0;x<radius;x++)
			{
				in[x+offset]=RombergIntegral(x-0.5,x+0.5,sd)*yint;
				totalsum=totalsum+in[x+offset];
			}
		}
		else
		{
			unsigned int ysqrd=y*y;
			for(unsigned int x=0;x<radius;x++)
			{
				in[x+offset]=1 / (2.0 * PI * sd * sd )* exp( (x * x + ysqrd)/(-2.0 * sd * sd ));
				totalsum=totalsum+in[x+offset];
			}
		}
		for(unsigned int x=width-1,xsrc=1;x>width-radius;x--,xsrc++)
		{
			in[x+offset]=in[xsrc+offset];
			totalsum=totalsum+in[x+offset];
		}
	}
	for(unsigned int y=height,srcoffset=width,offset=width*(height-1);y>height-radius;y--,offset-=width,srcoffset+=width)
	{
		for(unsigned int x=0;x<radius;x++)
		{
			in[x+offset]=in[x+srcoffset];
			totalsum=totalsum+in[x+offset];
		}
		for(unsigned int x=width-1,xsrc=1;x>width-radius;x--,xsrc++)
		{
			in[x+offset]=in[xsrc+srcoffset];
			totalsum=totalsum+in[x+offset];
		}
	}

	//const float k=0.10;
	float factor=/*k-*/1/totalsum;
	if(unsharp)
		factor=-strength*factor;

	for(unsigned int y=0,offset=0;y<radius;y++,offset+=width)
	{
		for(unsigned int x=0;x<radius;x++)
			in[x+offset]=in[x+offset]*factor;
		
		for(unsigned int x=width-1;x>width-radius;x--)
			in[x+offset]=in[x+offset]*factor;
		
	}
	for(unsigned int y=height,offset=width*(height-1);y>height-radius;y--,offset-=width)
	{
		for(unsigned int x=0;x<radius;x++)
			in[x+offset]=in[x+offset]*factor;
		for(unsigned int x=width-1;x>width-radius;x--)
			in[x+offset]=in[x+offset]*factor;
		
	}
	if(unsharp)
		*in=1.0+strength+*in;
	fftwf_execute(plan);
	//memcpy(gaus,out,outwidth*height*sizeof(complex<float>));
}


 /***********************************************************
  * Integral of a function gaus(X) by Romberg's method		*
  * ------------------------------------------------------- *
  * INPUTS:													*
  *          a      begin value of x variable				*
  *          b      end value of x variable					*
  *			sd      standard deviation						*
  *															*
  *															*
  * RETURNED VALUE  the integral of gausian from a to b		*
  *															*
  * REFERENCE: "Mathematiques en Turbo-Pascal (Part 1) by	*
  *             Marc Ducamp and Alain Reverchon, Eyrolles,	*
  *             Paris, 1987"								*
  *                             C++ version by J-P Moreau   *
  *            http://perso.wanadoo.fr/jean-pierre.moreau/  *
  *								    modified by t petersen  *
  **********************************************************/

  float GausBlur::RombergIntegral(float a,float b,float sd) 
  {
	#define FUNC(x) gausconst_a * exp( (x * x)/expdiv)
	float obtprec;
    int n;
	const float PREC=0.00001;
    const int ITERMIN=1;	
	const int ITERMAX=100;
	const float gausconst_a=1/(sqrt(2.0 * PI) * sd);
	const float expdiv= -2.0*sd*sd;

    int    i,j;
    float pas,r,s,ta;
    float t[ITERMAX][ITERMAX];
    r = FUNC(a);
    ta = (r + FUNC(b)) / 2;
    n=0;
    pas=b-a;
    t[0][0]=ta*pas;
	do {
      n = n + 1;
      pas = pas/2;
      s=ta;
	  for (i=1; i<pow(2.0,n); i++)
        s += FUNC(a+pas*i);
      t[0][n]=s*pas;
      r=1;
	  for (i=1; i<n+1; i++) {
        r=r*4;
        j=n-i;
        t[i][j]=(r*t[i-1][j+1] - t[i-1][j])/(r-1);
	  }
      obtprec = fabs(t[n][0] - t[n-1][0]);
      if (n>=ITERMAX) return t[n][0];
	} while (obtprec >PREC || n<ITERMIN); 
    return t[n][0];
#undef FUNC(x)
  }


void GausBlur::Blur(int plane,PVideoFrame &src,PVideoFrame &dst, IScriptEnvironment* env)
{
	/*
	LARGE_INTEGER timerB;
	LARGE_INTEGER timerA;
	QueryPerformanceCounter(&timerB);*/
	

	unsigned char* dstp=dst->GetWritePtr(plane);
	const unsigned char* srcp=src->GetReadPtr(plane);
	
	unsigned int dstpitch=dst->GetPitch(plane);
	unsigned int srcpitch=src->GetPitch(plane);

	for(unsigned int y=0,srcoffset=0,dstoffset=(border==1||border==2?0:radius*width+radius);y<srcheight;y++,srcoffset+=srcpitch,dstoffset+=width)
		for(unsigned int x=0;x<srcwidth;x++)
			in[x+dstoffset]=srcp[x+srcoffset];
	if(border==3)
	{
		//LeftBorder and right border 
		unsigned int srcoffset=0;
		unsigned int dstoffset=radius*width;
		for(unsigned int y=0;y<srcheight;y++,srcoffset+=srcpitch,dstoffset+=width){
			for(unsigned int x=0;x<radius;x++)
				in[dstoffset+x]=srcp[srcoffset];

			//memcpy(in+dstoffset+radius,srcp+srcoffset,srcwidth);

			for(unsigned int x=0;x<radius;x++)
				in[dstoffset+radius+srcwidth+x]=srcp[srcoffset+srcwidth-1];
		}
		//TopBorder;
		srcoffset=width*radius;
		dstoffset=0;
		for(unsigned int y=0;y<radius;y++,dstoffset+=width)
			memcpy(dstoffset+in,srcoffset+in,width*sizeof(float));
		//buttomborder;
		srcoffset=width*(radius+srcheight-1);
		dstoffset=width*(radius+srcheight);
		for(unsigned int y=0;y<radius;y++,dstoffset+=width)
			memcpy(dstoffset+in,srcoffset+in,width*sizeof(float));
	}


	fftwf_execute(plan);
	
	for(unsigned int y=0,offset=0;y<height;y++,offset+=outwidth)
		for(unsigned int x=0;x<(outwidth);x++)
			out[x+offset]*=gaus[x+offset];
	
	fftwf_execute(plani);

	
	if(border==2)
	{
		env->BitBlt(dstp,dstpitch,srcp,srcpitch,srcwidth,srcheight);
		for(unsigned int y=radius,srcoffset=radius*width,dstoffset=radius*dstpitch;y<srcheight-radius;y++,srcoffset+=width,dstoffset+=dstpitch)
			for(unsigned int x=radius;x<srcwidth-radius;x++)
			{
				float result=realout[x+srcoffset]*norm;
				dstp[x+dstoffset]=result<0?0:result>255?255:result;
			}
	}
	else
		for(unsigned int y=0,srcoffset=(border==1?0:radius*width+radius),dstoffset=0;y<srcheight;y++,srcoffset+=width,dstoffset+=dstpitch)
			for(unsigned int x=0;x<srcwidth;x++)
			{
				float result=realout[x+srcoffset]*norm;
				dstp[x+dstoffset]=result<0?0:result>255?255:result;
			}
	/*
QueryPerformanceCounter(&timerA);
	
	unsigned int time=timerA.LowPart-timerB.LowPart;
	char buf[100];
	itoa(time,buf,10);
	MessageBox(0,buf,"time",MB_OK);*/
			
}

